/* An abstract partial evaluator (ape), 
	using well-founded ordering framework developed by Bern */


%:- setdebug.

abint(Atom,Tree,Weighttable) :-
	canonical(Atom),
	get_weight(Atom,Weighttable,W,_),
	iterate([Atom],[[(Atom,W)]],Tree,Weighttable).

/* procedure iterate computes a sequence of successively greater sets
   of atoms (the pfg), terminating when two successive elements of the
   sequence are the same  */

iterate([],Tree,Tree,_) :-
	!,
	write('End of Analysis'),nl.
iterate(Newatoms,Tree,FinalTree,Weighttable) :-
	Newatoms \== [],
	trace_write(Newatoms),
	extend_pfg(Newatoms,Tree,NewTree,Weighttable,More),
	iterate(More,NewTree,FinalTree,Weighttable).

extend_pfg([],Tree,Tree,_,_) :-
	!.
extend_pfg([A|As],Tree,FinalTree,Weighttable,More) :-
	get_weight(A,Weighttable,AW,Atoms),
	alpha_insert(Atoms,(A,AW),Tree,NewTree,Weighttable,More),
	!,
	extend_pfg(As,NewTree,FinalTree,Weighttable,More).
	
/* alpha_insert:  updates the current Tree with the new atoms.
   If this results in a new atom then it is added to the list of new atoms.  */

alpha_insert([],_,Tree,Tree,_,_) :-
	!.
alpha_insert(As,B,[[B|Branch]|Tree],FinalTree,Weighttable,More) :-
	extend_branches(As,[B|Branch],FinalTree,RestTree,Weighttable,More),
	alpha_insert(As,B,Tree,RestTree,Weighttable,More).
alpha_insert(As,B,[[B1|Br]|Tree],[[B1|Br]|FinalTree], Weighttable,More) :-
	B \== B1,
	alpha_insert(As,B,Tree,FinalTree,Weighttable,More).
alpha_insert(_,_,[],[],_,_).

extend_branches([],_,T,T,_,_).
extend_branches([A|As],Branch,[[fail|Branch]|NewBranches],Tail,Weighttable,M) :-
	\+ hasweight(A,Weighttable),
	!,
	extend_branches(As,Branch,NewBranches,Tail,Weighttable,M).
extend_branches([A|As],Branch,[[loop,(C,CW)|Branch]|Brs],Tail,Weighttable,M) :-
	get_weight(A,Weighttable,W,_),
	branch_prefix([(C,CW)|_],Branch),
	leq(CW,W),
	lub(A,C,D),
	C = D,
	!,
	extend_branches(As,Branch,Brs,Tail,Weighttable,M).
extend_branches([A|As],Branch,[[(D,DW)|Pre]|Brs],Tail,Weighttable,M) :-
	get_weight(A,Weighttable,W,_),
	branch_prefix([(C,CW)|Pre],Branch),
	leq(CW,W),
	!,
	lub(A,C,D),
	C \== D,
	get_weight(D,Weighttable,DW,_),
	memb1(D,M),
	extend_branches(As,Branch,Brs,Tail,Weighttable,M).
extend_branches([A|As],Br,[[(A,W)|Br]|NewBranches],Tail,Weighttable,M) :-
	get_weight(A,Weighttable,W,_),
	memb1(A,M),
	extend_branches(As,Br,NewBranches,Tail,Weighttable,M).

hasweight(A,W) :-
	get_weight(A,W,_,_).

branch_prefix(B,B).
branch_prefix(P,[_|Bs]) :-
	branch_prefix(P,Bs).

/* The table of atoms and weight is stored as an ordered binary
   tree, with the atoms as keys.  */

get_weight(A,Weighttable,(Class,Weight),Atoms) :-
	var(Weighttable),
	!,
	argweights(A,Weight),
	unfold(A,Atoms,Class),
	Weighttable = tree(A,(Class,Weight),Atoms,_,_).
get_weight(A,tree(A,Weight,Atoms,_,_),Weight,Atoms) :-
	!.
get_weight(A,tree(A1,_,_,Left,_),Weight,Atoms) :-
	A @< A1,
	get_weight(A,Left,Weight,Atoms).
get_weight(A,tree(A1,_,_,_,Right),Weight,Atoms) :-
	A @> A1,
	get_weight(A,Right,Weight,Atoms).


lub(A,A1,A2) :-
	melt(A,MA),
	melt(A1,MA1),
	msg(MA,MA1,A2),
	canonical(A2).

insert_it(A,A,_,_) :-
	!.
insert_it(_,A,[A|_],More) :-
	memb1(A,More).

collect_atoms([],_,[]) :-
	!.
collect_atoms([(Weight,R)|Rs],As,Ps1) :-
	body_atoms(R,As),
	collect_atoms(Rs,As,Ps),
	insert_weight(Weight,Ps,Ps1).

insert_weight(Weight,P,P) :-
	memb1(Weight,P),
	!.
insert_weight(Weight,P,[Weight|P]).

body_atoms(G,_) :-
	var(G),
	!,
	nl,
	write('Warning: Variable goal encountered'),
	sp_ttyflush.
body_atoms(true,_) :-
	!.
body_atoms((B,Bs),As) :-
	!,
	body_atoms(B,As),
	body_atoms(Bs,As).
body_atoms(Bs,As) :-
	prolog_meta(Bs,_,P,_),
	!,
	body_atoms(P,As).
body_atoms(B,_) :-
	sp_builtin(B),
	!.
body_atoms(B,_) :-
	imported(B),
	!.
body_atoms(B,As) :-
	copyterm(B,A),
	canonical(A),
	memb1(A,As),
	!.
body_atoms(_,_).


unfold(A,Atoms,P/N) :-
	trace_atom(A),
	melt(A,Ma),
	bagof((Weight,R), Ma^
		resultant(Ma,Weight,R),
		Rs),
	functor(A,P,N),
	collect_atoms(Rs,Atoms,_).

:- dynamic trace_on/0.
:- dynamic atom_trace/0.

/* trace_on. */

trace_atom(X) :-
	atom_trace,
	!,
	nl,write(X),sp_ttyflush.
trace_atom(_) :-
	write('.'),
%	drip,
	sp_ttyflush.

trace_write(X) :-
	trace_on,
	!,
	nl,write(X),nl.
trace_write(_) :-
	write('*'),sp_ttyflush.

% argweights(_,1).
argweights(T,W) :-
	select_args(T,As),
	!,
	weightlist(As,Ws),
	sumlist(Ws,W).
argweights(T,W) :-
	weight(T,W).

select_args(T,As) :-
	functor(T,P,N),
	functor(T1,P,N),
	ub_clause(measure(T1),true),
	T =.. [P|Xs],
	T1 =.. [P|Ys],
	extract_weighted_args(Xs,Ys,As).

extract_weighted_args([],[],[]).
extract_weighted_args([X|Xs],[1|Ys],[X|Zs]) :-
	extract_weighted_args(Xs,Ys,Zs).
extract_weighted_args([_|Xs],[0|Ys],Zs) :-
	extract_weighted_args(Xs,Ys,Zs).

weight(X,1) :-
	var(X),
	!.
weight(X,N) :-
	X =.. [_|Xs],
	weightlist(Xs,Ws),
	sumlist([1|Ws],N).

weightlist([],[]).
weightlist([X|Xs],[N|Ns]) :-
	weight(X,N),
	weightlist(Xs,Ns).

sumlist([],0).
sumlist([X|Xs],Y) :-
	sumlist(Xs,Y1),
	Y is Y1+X .


lessthan((C,X),(C,Y)) :- 
	X < Y.

leq(X,X) :-
	!.
leq(X,Y) :-
	lessthan(X,Y).
