% compute closed convex hull of a given set of convex polygons.
% use linear relaxation

% e.g. convhull([X,Y],[(X > 0, X=Y),(X+Y=0)],Hull,Coords).

% Answer:  Hull = [_A+_B>=0], Coords = [_A,_B] ?


convhull(Coords,Cs,Hull,Ys) :-
	convex_hull(Coords,Cs,Xs),
	dump(Xs,Ys,Hull1),
	anti_alias(Xs,Ys,Hull1,Hull).

anti_alias(Xs,Ys,Hull1,Hull) :-
	makepairs(Xs,Ys,Ps),
	extract_aliases(Ps,Hull,Hull1).

makepairs([],[],[]).
makepairs([X|Xs],[Y|Ys],[(X/Y)|Ps]) :-
	makepairs(Xs,Ys,Ps).

extract_aliases([(X/Y)|Ps],E0,E2) :-
	find_aliases(X,Y,Ps,E0,E1),
	extract_aliases(Ps,E1,E2).
extract_aliases([],Es,Es).

find_aliases(X,Y,[(X1/Z)|Ps],[Y=Z|Es],Es1) :-
	X == X1,
	!,
	find_aliases(X,Y,Ps,Es,Es1).
find_aliases(X,Y,[_|Ps],Es,Es1) :-
	find_aliases(X,Y,Ps,Es,Es1).
find_aliases(_,_,[],Es,Es).

convex_hull(Points,[C],Xs) :-
	!,	% for a single constraint just return it.
	copyterm((Points,C),(Xs,C1)),
	{C1}.
convex_hull(Points, Constraints,Xs) :-
	lambda_subst(Points, Constraints, Lambdas, NewCoords, Cs0,Cs1),
	conv_comb(NewCoords, Xs, Cs1, Cs2),
	polytope(Lambdas,Cs2,[]),
	constraints(Cs0).

polytope(Xs,Cs0,Cs2) :-
	positive_sum(Xs, 1, Cs0, Cs2).

positive_sum([], Z, [Z=0|Cs],Cs).

positive_sum([X|Xs], SumX, [X >= 0, SumX = X+Sum | Cs0],Cs1) :-
	positive_sum(Xs, Sum, Cs0,Cs1).

conv_comb(Coords,Xs,Cs1,Cs2) :-
	transpose(Coords,Coords1),
	conv_comb1(Coords1, Xs, Cs1, Cs2).

conv_comb1([],[],Cs,Cs).
conv_comb1([Ps|Rest], [X|Xs],[X = Sum |Cs0],Cs1) :-
	sumpoints(Ps, Sum),
	conv_comb1(Rest,Xs,Cs0,Cs1).

sumpoints([],0) :-
	!.
sumpoints([X],X) :-
	!.
sumpoints([X1,X2|Xs],X1+Xs1) :-
	sumpoints([X2|Xs],Xs1).

transpose(Xs,[]) :-
	nullrows(Xs).
transpose(Xs,[Y|Ys]) :-
	makerow(Xs,Y,Zs),
	transpose(Zs,Ys).

makerow([],[],[]).
makerow([[X|Xs]|Ys],[X|Xs1],[Xs|Zs]):-
	makerow(Ys,Xs1,Zs).

nullrows([]).
nullrows([[]|Ns]) :-
	nullrows(Ns).


constraints([]).
constraints([true|Cs]) :-
	!,
	constraints(Cs).
constraints([C|Cs]) :-
	{C},
	constraints(Cs).

lin_comb_r([],[],[],Cs,Cs).
lin_comb_r([P|Ps], [S|Ss], [Kps|Ss1],[Kps = P+S | Cs], Cs1) :-
	lin_comb_r(Ps,  Ss, Ss1, Cs, Cs1).

lambda_subst(_, [], [], [], Cs, Cs).
lambda_subst(Ps, [C|Cs], [L|Ls], [Ws|Newpoints], [C1|Cs1],Cs2) :-
	lambda_dict(Ps,Ws,D),
	lambda_subst_constraint(C,L,C1,D),
	lambda_subst(Ps,Cs,Ls,Newpoints,Cs1,Cs2).

lambda_dict([],[],[]).
lambda_dict([P|Ps],[W|Ws],[subst(P,W)|D]) :-
	lambda_dict(Ps,Ws,D).

% Multiply both sides of each inequality by L, and for
% each variable X replace X*L by a new variable.
% note that since L could be zero, multiplying each side by L
% changes a strict inequality into a loose inequality.

lambda_subst_constraint((C,Cs),L,(C1,Cs1),D) :-
	lambda_subst_constraint(C,L,C1,D),
	lambda_subst_constraint(Cs,L,Cs1,D).
lambda_subst_constraint(E1=E2,L,F1=F2,D) :-
	lambda_subst_term(E1,L,F1,D),
	lambda_subst_term(E2,L,F2,D).
lambda_subst_constraint(E1=:=E2,L,F1=:=F2,D) :-
	lambda_subst_term(E1,L,F1,D),
	lambda_subst_term(E2,L,F2,D).
lambda_subst_constraint(_=\=_,_,true,_).
lambda_subst_constraint(E1=<E2,L,F1=<F2,D) :-
	lambda_subst_term(E1,L,F1,D),
	lambda_subst_term(E2,L,F2,D).
lambda_subst_constraint(E1>=E2,L,F1>=F2,D) :-
	lambda_subst_term(E1,L,F1,D),
	lambda_subst_term(E2,L,F2,D).
lambda_subst_constraint(E1>E2,L,F1>=F2,D) :-	
	lambda_subst_term(E1,L,F1,D),
	lambda_subst_term(E2,L,F2,D).
lambda_subst_constraint(E1<E2,L,F1=<F2,D) :-	
	lambda_subst_term(E1,L,F1,D),
	lambda_subst_term(E2,L,F2,D).
lambda_subst_constraint(true,_,true,_).

lambda_subst_term(X,_,W,D) :-
	var(X),
	!,
	lookup_lambda_dict(X,W,D).
lambda_subst_term(rat(X,Y),L,(X*L)/Y,_) :-
	!.
lambda_subst_term(X1,L,L*X1,_) :-
	number(X1),
	!.
lambda_subst_term(X1+X2,L,Y1+Y2,D) :-
	lambda_subst_term(X1,L,Y1,D),
	lambda_subst_term(X2,L,Y2,D).
lambda_subst_term(X1-X2,L,Y1-Y2,D) :-
	lambda_subst_term(X1,L,Y1,D),
	lambda_subst_term(X2,L,Y2,D).
lambda_subst_term(+X1,L,+Y1,D) :-
	lambda_subst_term(X1,L,Y1,D).
lambda_subst_term(-X1,L,-Y1,D) :-
	lambda_subst_term(X1,L,Y1,D).
lambda_subst_term(X1*X2, L, Y1*X2, D) :-
	ground(X2),
	!,
	lambda_subst_term(X1, L, Y1, D).
lambda_subst_term(X1*X2, L, X1*Y2, D) :-
	!,
	lambda_subst_term(X2, L, Y2, D).
lambda_subst_term(X1/X2, L, Y1/X2, D) :-	% note X2 assumed to be ground
	ground(X2),
	!,
	lambda_subst_term(X1, L, Y1, D).
	

lookup_lambda_dict(X,W,[subst(X1,W)|_]) :-
	X == X1,
	!.
lookup_lambda_dict(X,W,[_|D]) :-
	lookup_lambda_dict(X,W,D).
