/* An abstract partial evaluator (ape), loosely based on the pe algorithm of
  Benkerimi and Lloyd, and using the characteristic path 
  abstraction and renaming techniques of Gallagher and Bruynooghe. 

  Fully described in Gallagher, J.P.; A System for Specialising Logic
  Programs;  Tech. Report CSTR-91-32; Department of Computer Science,
  University of Bristol */

%:- setdebug.

abint(Atom,Pfg,Pathtable) :-
	canonical(Atom),
	alpha_insert([Atom],Pfg,Pathtable,More),
	iterate(More,Pfg,Pathtable).

/* procedure iterate computes a sequence of successively greater sets
   of atoms (the pfg), terminating when two successive elements of the
   sequence are the same  */

iterate([],_,_) :-
	write('End of Analysis'),nl.
iterate(Newatoms,Atomset,Pathtable) :-
	Newatoms \== [],
	trace_write(Newatoms),
	extend_pfg(Newatoms,Atomset,Pathtable,More),
	iterate(More,Atomset,Pathtable).

extend_pfg([],_,_,_) :-
extend_pfg([A|As],Atomset,Pathtable,More) :-
	get_path(A,Pathtable,_,Atoms),
	alpha_insert(Atoms,Atomset,Pathtable,More),
	extend_pfg(As,Atomset,Pathtable,More).
	
/* alpha_insert:  updates the current Atomset with the new atoms.
   The most specific generalisation of each set of atoms with the
   same path is computed.  If this results in a new atom then it is
   added to the list of new atoms.  */

alpha_insert([],_,_,_) :-
alpha_insert([A|As],Atomset,Pathtable,More) :-
	get_path(A,Pathtable,Path,_),
	!,
	insert_atom(Path,A,Atomset,More),
	alpha_insert(As,Atomset,Pathtable,More).
alpha_insert([_|As],Atomset,Pathtable,More) :- /* atom fails - no path */
	alpha_insert(As,Atomset,Pathtable,More).
	

/* The table of atoms and paths is stored as an ordered binary
   tree, with the atoms as keys.  */

get_path(A,Tree,Path,Atoms) :-
	var(Tree),
	!,
	unfold(A,Rs,Path),
	collect_atoms(Rs,Atoms,[]),
	Tree = tree(A,Path,Atoms,_,_).
get_path(A,tree(A,Path,Atoms,_,_),Path,Atoms) :-
get_path(A,tree(A1,_,_,Left,_),Path,Atoms) :-
	A @< A1,
	get_path(A,Left,Path,Atoms).
get_path(A,tree(A1,_,_,_,Right),Path,Atoms) :-
	A @> A1,
	get_path(A,Right,Path,Atoms).

/* insert_atom(Path,A,Atomset,More):  insert the atom A with Path
   into the Atomset.   Find the msg of A and the set of atoms in
   Atomset with the same Path.  If the result is not in Atomset, then
   insert it into Atomset and add it to the list More. */

insert_atom(Path,A,Atomset,More) :-
	var(Atomset),
	!,
	Atomset = [element(Path,[A|_])|_],
	memb1(A,More).
insert_atom(Path,A,[element(P1,As)|_],More) :-
	Path == P1,
	abstract(A,As,More).
insert_atom(Path,A,[element(P1,_)|Atomset],More) :-
	Path \== P1,
	insert_atom(Path,A,Atomset,More).

abstract(A,[A1|As],More) :-
	var(As),
	!,
	lub(A,A1,A2),
	insert_it(A1,A2,As,More).
abstract(A,[_|As],More) :-
	abstract(A,As,More).

lub(A,A1,A2) :-
	melt(A,MA),
	melt(A1,MA1),
	msg(MA,MA1,A2),
	canonical(A2).

insert_it(A,A,_,_) :-
	!.
insert_it(_,A,[A|_],More) :-
	memb1(A,More).

collect_atoms([],As,As) :-
collect_atoms([R|Rs],As,As2) :-
	body_atoms(R,As,As1),
	collect_atoms(Rs,As1,As2).

resultants_paths([],[],[]).
resultants_paths([(Path,R)|Rs],[R|Rs1],Ps1) :-
	canonical(R),
	resultants_paths(Rs,Rs1,Ps),
	insert_path(Path,Ps,Ps1).

insert_path(Path,P,P) :-
	memb1(Path,P),
	!.
insert_path(Path,P,[Path|P]).

body_atoms(G,As,As) :-
	variable(G),
	nl,
	write('Warning: Variable goal encountered'),
	sp_ttyflush.
body_atoms(true,As,As) :-
	!.
body_atoms((B,Bs),As,As2) :-
	!,
	body_atoms(B,As,As1),
	body_atoms(Bs,As1,As2).
body_atoms(Bs,As,As1) :-
	prolog_meta(Bs,_,P,_),
	!,
	body_atoms(P,As,As1).
body_atoms(B,Q,Q) :-
	sp_builtin(B),
	!.
body_atoms(B,Q,Q) :-
	imported(B),
	!.
body_atoms(B,[A|Q],Q) :-
	copyterm(B,A),
	canonical(A),
	!.
body_atoms(_,Q,Q).


unfold(A,Rs,P) :-
	trace_atom(A),
	melt(A,Ma),
	bagof((Path,R), Ma^
		resultant(Ma,Path,R),
		RPs),
	resultants_paths(RPs,Rs,P).


:- dynamic trace_on/0.
:- dynamic atom_trace/0.

/* trace_on. */

trace_atom(X) :-
	atom_trace,
	!,
	nl,write(X),sp_ttyflush.
trace_atom(_) :-
	write('.'),
%	drip,
	sp_ttyflush.

trace_write(X) :-
	trace_on,
	!,
	nl,write(X),nl.
trace_write(_) :-
	write('*'),sp_ttyflush.
