% JPG - 14.2.92
% Compute a regular approximation of a given program.
% Based on
% Gallagher and de Waal:  "Regular Approximations of
% Logic Programs and Their Uses";  CSTR-92-06.
% with ideas from "Logic Programs as Types for Logic Programs"
% by Fruehwirth, Shapiro, Vardi, Yardeni

%:- setdebug.
 
%:- alldynamic.

tp_fix(P) :-
	time(tp_fix1(P,F)),
	display_fix(F).
	
tp_fix1(P,F) :-
	strong_connected_components(P,G),% compute the dependency ordering
	nl,
	tp_fix_groups(G,[],F,10,_).

tp_fix_groups([],F,F,K,K) :-
	!.
tp_fix_groups([(R,G)|Gs],F,F2,K,K2) :-
	nonrec(R),
	write(G),nl,
	!,
	init(G,R0),
	approx_names(G,Gapps),
	tp_alpha(G,Gapps,[[]],F,R0,R2,K,K1),  % no need to iterate if non-rec.
	app(F,R2,F1),
	fold(F1,F3),
	tp_fix_groups(Gs,F3,F2,K1,K2).
tp_fix_groups([(_,G)|Gs],F,F2,K,K2) :-
	write(G),nl,
	init(G,R0),
	tp_iterate([[]],G,F,R0,R2,K,K1),
	app(F,R2,F1),
	tp_fix_groups(Gs,F1,F2,K1,K2).

nonrec(non_recursive).

recursive_pred(X) :-
	memb2(P,X),
	recursive(P),
	!.

init([_|Ps],[[]|R]) :-
	init(Ps,R).
init([],[]).

tp_iterate([],_,_,R,R,K,K) :-
	!.
tp_iterate(Nofix,Ps,Acc,R1,R,K,K2) :-
	writestate(R1),
	app(R1,Acc,D),
	approx_names(Ps,Papps),
	tp_alpha(Ps,Papps,Nofix,D,R1,R2,K,K1),
	fold(R2,R3),
	countdefs(R3),
	check_fix(Ps,R1,R3,Fix),
	!,
	tp_iterate(Fix,Ps,Acc,R3,R,K1,K2).

approx_names([],[]).
approx_names([P/_|Ps],[Q|Qs]) :-
	name(P,Fn),
        name('_x',Suff),
        app(Fn,Suff,Fapp),
        name(Q,Fapp),
	!,
	approx_names(Ps,Qs).

countdefs([R|Rs]) :-
	sp_length(R,N),write(N),write(' '),
	!,
	countdefs(Rs).
countdefs([]) :-
	nl.

	
tp_alpha([],[],_,_,[],[],K,K).
tp_alpha([P/N|Ps],[Papprox|Papps],Nofix,R,[PR|PRs],[UP|R1],K,K2) :-
	user_clauses(P/N,Cls),
	infer_proc(Cls,Nofix,R,As,Papprox,K,K1),
	regularise_proc(Papprox,P/N,UP1,As,PR,Change),
	rdefs2(Change,Papprox,UP,UP1,[]),
	tp_alpha(Ps,Papps,Nofix,R,PRs,R1,K1,K2).

regularise_proc(_,_,P,[],P,nochange).
regularise_proc(A,Q,U,[A1|As],P,change) :-	
	regularise(A,Q,U1,A1,P),
	regularise_proc1(A,Q,U,[U1|As]).

regularise_proc1(_,_,U,[U]).
regularise_proc1(A,Q,U,[A1,A2|As]) :-
	regularise(A,Q,U1,A1,A2),
	regularise_proc1(A,Q,U,[U1|As]).

rdefs2(change,P,R,D,Acc) :-
	rdefs(P,R,D,Acc).
rdefs2(nochange,_,D,D,_).

	
infer_proc([(_ :- B)|Cls],Nofix,R,As,P,K,K1) :- % if no predicate in the body
	nochange(B,Nofix),			% changed, skip clause
	!,
	infer_proc(Cls,Nofix,R,As,P,K,K1).
infer_proc([(H :- B)|Cls],Nofix,R,[A2|As],P,K,K3) :-
	tsolve(B,R,T,Bdefs),
	flatten_defs(Bdefs,[],Fdefs),
	unfold_body(T,Fdefs,T1),
	intersect_body(T1,Fdefs,T2,S,K,K1),
	!,
	headtype(H,T2,P,A,K1,K2), 	
	app(A,S,A1),
	rdefs(P,A2,A1,Fdefs),
	infer_proc(Cls,Nofix,R,As,P,K2,K3).
infer_proc([_|Cls],Nofix,R,As,P,K,K1) :-		% if no solution in R.
	infer_proc(Cls,Nofix,R,As,P,K,K1).
infer_proc([],_,_,[],_,K,K).

nochange(_,[[]]) :-  	% everything handled on 1st iteration
	!,
	fail.
nochange(true,_) :-
	!.
nochange(B,Nofix) :-
	preds(B,Ps),
	\+ commonpred(Ps,Nofix).

preds(true,[]) :-
	!.
preds((B,Bs), [P/N|Ps]) :-
	!,
	functor(B,P,N),
        preds(Bs,Ps).
preds(B,[P/N]) :-
	functor(B,P,N).

commonpred(X,Y) :-
	memb2(Z,X),
	memb2(Z,Y).
	
tsolve(true,_,true,[]) :-
	!.
tsolve((B,Bs),R,T,[D|Ds]) :-
	!,
	tsolve(B,R,T1,D),
	tsolve(Bs,R,T2,Ds),
	joingoals(T1,T2,T).
tsolve(B,_,Bs,[]) :-
        sp_builtin(B),
        !,
        any_term(B,Bs).
tsolve(B,R,T1,Bdef) :-
	functor(B,S,N),
	headclause(S/N,R,(H :- T),Bdef),
	melt((H :- T),(H1 :- T1)),
	H1 =.. [_,B].

unfold_body(true,_,true) :-
	!.
unfold_body((B,Bs),R,C) :-
	!,
	unfold_body(B,R,C1),
	unfold_body(Bs,R,C2),
	joingoals(C1,C2,C).
unfold_body(B,_,B) :-
	B =.. [_,X],
	var(X),
	!.
unfold_body(B,_,Bs) :-
	B =.. [any,_],
	!,
	any_term(B,Bs).
unfold_body(B,R,B2) :-
	B =.. [T,_],
	typeclause(T/1,R,(H :- B1)),
	melt((H :- B1),(B :- B3)),
	unfold_body(B3,R,B2).

any_term(B,Bs) :-
	vars(B,Vs),
	any_vars(Vs,Bs).

any_vars([],true) :-
	!.
any_vars([X],any(X)) :-
	!.
any_vars([X|Xs],(any(X),Bs)) :-
	any_vars(Xs,Bs).
	
intersect_body(true,_,true,[],K,K) :-
	!.
intersect_body((B,Bs),R,C,I,K,K2) :-
	reoccurs(B,Bs,B1,Bs1),
	!,
	B =.. [T1,X],
	B1 =.. [T2,X],
	intersect(T1,T2,T3,P,R,K,K1),
	B2 =.. [T3,X],
	app_sorted_defs(P,R,PR),
        intersect_body((B2,Bs1),PR,C,J,K1,K2),
	app(P,J,I).
intersect_body((B,Bs),R,(B,Bs1),I,K,K1) :-
	!,
	intersect_body(Bs,R,Bs1,I,K,K1).
intersect_body(B,_,B,[],K,K).
	
	
reoccurs(B,(B1,Bs),B1,Bs) :-
	B =.. [_,X],
	B1 =.. [_,Y],
	X == Y,
	!.
reoccurs(B,(B2,Bs),B1,(B2,Bs1)) :-
	!,
	reoccurs(B,Bs,B1,Bs1).
reoccurs(B,B1,B1,true) :-
	B =.. [_,X],
	B1 =.. [_,Y],
	X == Y,
	!.
	
headtype(T, Body,P,[proc(P/1,[(S :- B)])|Argtypes],K,K2) :- 
	functor(T,F,N),
	functor(X,F,N),
	name(F,Fn),
	S =.. [P,X],
	canonical(S),
	argtypenames(Fn,N,1,Ns),
	top_argtypes(T,Ns,Body,X,B,1,N,K,K2,Argtypes).
	
top_argtypes(_,_,_,_,true,_,0,K,K,[]).
top_argtypes(T,[J],Body,X,B,N,N,K,K1,Argtypes) :-
	top_argtype(T,J,Body,X,B,N,K,K1,Argtypes).
top_argtypes(T,[J|Js],Body,X,(B,Bs),M,N,K,K2,Argtypes) :-
	M < N,
	top_argtype(T,J,Body,X,B,M,K,K1,Argtype),
	M1 is M+1,
	top_argtypes(T,Js,Body,X,Bs,M1,N,K1,K2,Argtypes1),
	app(Argtype,Argtypes1,Argtypes). 
	
top_argtype(T,_,Body,X,B,M,K,K,[]) :-
	arg(M,T,Xm),
	var(Xm),
	!,
	arg(M,X,Ym),
	giventype(Xm,Body,R),
	B =.. [R,Ym].
top_argtype(T,J,Body,X,B,M,K,K1,Types) :-
	arg(M,T,Xm),
	arg(M,X,Ym),
	top_termtype(Xm,J,Body,Types,K,K1),
	B =.. [J,Ym].

top_termtype(T, J, Body,[proc(J/1,[(S :- B)])|Argtypes],K,K1) :- 
	functor(T,F,N),
	functor(X,F,N),
	S =.. [J,X],
	canonical(S),
	argtypes(T,Body,X,B,1,N,K,K1,Argtypes).

termtype(T, Body,[proc(P/1,[(S :- B)])|Argtypes],K,K2) :- 
	functor(T,F,N),
	functor(X,F,N),
	newname(K,P),
	S =.. [P,X],
	canonical(S),
	K1 is K+1,
	argtypes(T,Body,X,B,1,N,K1,K2,Argtypes).

argtypes(_,_,_,true,_,0,K,K,[]).
argtypes(T,Body,X,B,N,N,K,K1,Argtypes) :-
	argtype(T,Body,X,B,N,K,K1,Argtypes).
argtypes(T,Body,X,(B,Bs),M,N,K,K2,Argtypes) :-
	M < N,
	argtype(T,Body,X,B,M,K,K1,Argtype),
	M1 is M+1,
	argtypes(T,Body,X,Bs,M1,N,K1,K2,Argtypes1),
	app(Argtype,Argtypes1,Argtypes). 
	
argtype(T,Body,X,B,M,K,K,[]) :-
	arg(M,T,Xm),
	var(Xm),
	!,
	arg(M,X,Ym),
	giventype(Xm,Body,R),
	B =.. [R,Ym].
argtype(T,Body,X,B,M,K,K1,Types) :-
	arg(M,T,Xm),
	arg(M,X,Ym),
	termtype(Xm,Body,Types,K,K1),
	newname(K,P),
	B =.. [P,Ym].

giventype(X,(B,_),T) :-
	B =.. [T,Y],
	X == Y,
	!.
giventype(X,(_,Bs),T) :-
	!,
	giventype(X,Bs,T).
giventype(X,B,T) :-
	B =.. [T,Y],
	X == Y,
	!.
giventype(_,_,any).
	
newname(K,P) :-
	name(K,Kn),
	name(t,T),
	app(T,Kn,Tn),
	name(P,Tn).

check_fix([],[],[],[]).
check_fix([_|Ps],[RP|R],[RP1|S],Qs) :-
        fixpoint(RP,RP1),
        !,
        check_fix(Ps,R,S,Qs).
check_fix([P|Ps],[_|R],[_|S],[P|Qs]) :-
        check_fix(Ps,R,S,Qs).

fixpoint([],[]) :-
        !.
fixpoint(R1,R2) :-
        toptype(R1,T1),
        toptype(R2,T2),
        subtype(T2,T1,R2,R1).


toptype([proc(T1/1,_)|_],T1).


headclause(P/N,R,C,Xdef) :-
	functor(X,P,N),
	hclause(X,R,C,Xdef).

hclause(H,[[proc(T/1,[(TH :- B)])|Hdef]|_],(TH :- B),Hdef) :-
	TH =.. [T,H],
	!.
hclause(P,[_|Ps],Cl,Def) :-
	hclause(P,Ps,Cl,Def).


nested_memb(X,[X|_]).
nested_memb(X,[Y|_]) :-
	nested_memb(X,Y).
nested_memb(X,[_|Z]) :-
	nested_memb(X,Z).


typeclause(T/N,P,Cl) :-
	sorteddef(T/N,Q,P),
	!,
	memb2(Cl,Q).
	
typeclause1(T/N,P,Cl) :-
	def(T/N,Q,P),
	!,
	memb2(Cl,Q).
	


:- dynamic traceon/0.

writestate(R) :-
	traceon,
	write(R),
	read(_),
	!,
	nl.
writestate(_) :-
	write('*'),
	sp_ttyflush.

lit(B, (B,_)).
lit(B, (_,Bs)) :-
        !,
        lit(B,Bs).
lit(B,B).

display_fix(Cls) :-
        nl,write('Enter filename: '),
        read(F),
        reg_filewrite(F,Cls).

reg_filewrite(tty,Cls) :-
        !,
	nl,write('Regular Approximation Program:'),nl,nl,
        displayqTypes(Cls),
	nl,nl.
reg_filewrite(F,Cls) :-
        tell(F),
        displayqTypes(Cls),
        told.

/* write the program to current output file  */

displayqTypes([[]|Ps]) :-
	displayqTypes(Ps).
displayqTypes([[Q|Qs]|Ps]) :-
	displayqPreddef(Q),
	displayqTypes(Qs),
	displayqTypes(Ps).
displayqTypes([proc(_,Qs)|Ps]) :-
	displayqRClauses(Qs),
	displayqTypes(Ps).
displayqTypes([]).

displayqPreddef(proc(_,[(H :- B)])) :-
	H =.. [_,A],
	displayqRClauses([(A :- B)]).


argtypenames(_,M,N,[]) :-
	N > M,
	!.
argtypenames(Pn,K,M,[N|Ns]) :-
	name(M,Mn),
	my_append(Pn,[95|Mn],PMn),
	name(N,PMn),
	M1 is M+1,
	argtypenames(Pn,K,M1,Ns).


fold([],[]).
fold([RP|R1],[RP1|R3]) :-
	foldp(RP,RP1),
	fold(R1,R3).
	

foldp(RP,RP3) :-
	depends(T1,T2,TPs,RP),
	same_functor_structure(T2,T1,TPs),
	recursive(T1,T2,TPs,RP,RP1),
	!,
	toptype(RP1,T),
	rdefs(T,RP2,RP1,[]),
	foldp(RP2,RP3).
foldp(R,R).

recursive(T1,T2,TPs,RP,RP2) :-
	subtype(T2,T1,TPs),
	!,
	recursive1(T1,T2,TPs,RP,RP2).

recursive1(T1,T2,TPs,RP,RP2) :-
	subtype(T1,T2,TPs),
	!,
	normalise_defs(RP,T2,T1,RP2).
	
depend_path([proc(T1/1,Cls)|Ps],T1,Path) :-
	body_pred(Cls,T),
	T1 \== T,
	T \== any,
	dep_path_cls(Ps,T,Path,[T1]).

dep_path_cls(_,T,[T|D],D).
dep_path_cls(TPs,T,Path,D) :-
	def(T/1,Cls,TPs),
	body_pred(Cls,T1),
	\+ memb2(T1,[T|D]),
	T1 \== any,
	dep_path_cls(TPs,T1,Path,[T|D]).
	
body_pred([(_ :- B)|_],T) :-
        bodytype(B,T).
body_pred([_|Cs],T) :-
        body_pred(Cs,T).

bodytype((B,_),T) :-
        functor(B,T,1).
bodytype((_,Bs),T) :-
	!,
        bodytype(Bs,T).
bodytype(B,T) :-
        functor(B,T,1).


	
replace_procs([],R,R).
replace_procs([P|Ps],R,R1) :-
	replace_proc(P,R,R2),
	replace_procs(Ps,R2,R1).

replace_proc(proc(T1/1,Cls),[proc(T1/1,_)|R],[proc(T1/1,Cls)|R]) :-
	!.
replace_proc(P,[P1|R],[P1|R1]) :-
	replace_proc(P,R,R1).
replace_proc(P,[],[P]).


less_functors(T1,T2,R) :-
	\+ morefunctors(T1,T2,R).

morefunctors(T1,T2,R) :-
	headfunctor(F,T1,R),
	\+ headfunctor(F,T2,R).

same_functor_structure(T1,T2,RP) :-
	\+ morefunctors(T1,T2,RP),
	\+ morefunctors(T2,T1,RP).



headfunctor(F/N,T,Defs) :-
        typeclause1(T/1,Defs,(H :- _)),
        H =.. [T,X],
        functor(X,F,N).

	
depends(T1,T2,R1,[_|P]) :-
	nested_memb(proc(T1/1,_),P),
	iterate_defs([T1],[T1|S],R1,[],P,[]),
	makelist(S),
	memb2(T2,S).

makelist([]) :-
        !.
makelist([_|X]) :-
        makelist(X).


flatten_defs([],D,D).
flatten_defs([proc(T/1,C)|Ds],D,D1) :-
	!,
	insert_def(T/1,C,D,D2),
	flatten_defs(Ds,D2,D1).
flatten_defs([P|Ps],D,D1) :-
	flatten_defs(P,D,D2),
	flatten_defs(Ps,D2,D1).

/*
insert_def(T,C,[],[proc(T,C)]) :-
	!.
insert_def(T,_,[proc(T,C1)|Ds],[proc(T,C1)|Ds]) :-
	!.
insert_def(T,C,[proc(T1,C1)|Ds],[proc(T1,C1)|Ds1]) :-
	T1 @< T,
	!,
	insert_def(T,C,Ds,Ds1).
insert_def(T,C,[proc(T1,C1)|Ds],[proc(T,C),proc(T1,C1)|Ds]) :-
	T @< T1.
*/

insert_def(T,C,[],[proc(T,C)]) :-
	!.
insert_def(T,_,[proc(T,C1)|Ds],[proc(T,C1)|Ds]) :-
	!.
insert_def(T,C,[proc(T1,C1)|Ds],[proc(T1,C1)|Ds1]) :-
	insert_def(T,C,Ds,Ds1).

app_sorted_defs([],P,P).
app_sorted_defs([proc(T,D)|Ds],P,P1) :-
	insert_def(T,D,P,P2),
	app_sorted_defs(Ds,P2,P1).
